@@ -46,6 +46,12 @@ arma::vec reRate(const arma::vec& T,
       --indices_head;
     w_sum += W(idx_i);
     de(idx_i) = w_sum;
+    // adjust for ties
+    for(arma::uword j = 1; j <= i; ++j) {
+      if (T[idx_i] == T(idx[i - j]))
+				de(idx[i - j]) = w_sum;
+      else break;
+    }   
   }
   for (arma::uword k = 0; k < m; k++) {
     for (arma::uword i = 0; i < n; i++) {
@@ -91,14 +97,22 @@ arma::mat matvec(arma::mat x, arma::vec y) {
   return out;
 }
 
+arma::mat matvec2(arma::mat x, arma::vec y) {
+  arma::mat out(x.n_rows, x.n_cols);
+  for (size_t i = 0; i < x.n_rows; i++) {
+    out.row(i) = x.row(i) % y.t();
+  }
+  return out;
+}
+
 // Log-rank estimating equation 
 //' @noRd
 // [[Rcpp::export(rng = false)]]
 arma::vec reLog(const arma::vec& a,
-		const arma::mat& X,
-		const arma::vec& T,
-		const arma::vec& Y,
-		const arma::vec& W) {
+								const arma::mat& X,
+								const arma::vec& T,
+								const arma::vec& Y,
+								const arma::vec& W) {
   arma::uword const n = Y.n_elem;
   arma::uword const p = a.n_elem;
   arma::vec out(p, arma::fill::zeros);
@@ -107,16 +121,19 @@ arma::vec reLog(const arma::vec& a,
   arma::vec yexa = log(Y) + X.t() * a;
   arma::uvec const idx = arma::sort_index(texa);
   auto cmp = [](cmp_par const &x, cmp_par const &y){
-	       return x.first <= y.first;
-	     };
+		return x.first <= y.first;
+	};
   std::set<cmp_par, decltype(cmp)> indices(cmp);
   double w_sum{};
   arma::vec x_col_sum(p, arma::fill::zeros);
+  arma::vec diff0(p, arma::fill::zeros);
+  arma::vec diff1(p, arma::fill::zeros);
   {
     auto const idx_i = idx[0];
     indices.emplace(yexa[idx_i], idx_i);
     x_col_sum += W(idx_i) * X.col(idx_i);
     w_sum += W(idx_i);
+		diff0 = -x_col_sum / w_sum;
   }
   auto indices_head = indices.begin();
   for(arma::uword i = 1; i < n; ++i) {
@@ -132,8 +149,18 @@ arma::vec reLog(const arma::vec& a,
       --indices_head;
     x_col_sum += W(idx_i) * X.col(idx_i);
     w_sum += W(idx_i);
-    if (w_sum > 0) 
-      out += W(idx_i) * (X.col(idx_i) - x_col_sum / w_sum);
+		if (w_sum > 0) {
+      diff1 = -x_col_sum / w_sum;
+      out += W(idx_i) * (X.col(idx_i) + diff1);
+      // adjust for ties
+      for(arma::uword j = 1; j <= i; ++j) {
+				auto const idx_j = idx[i - j];
+				if (texa[idx_i] == texa[idx_j])
+					out += W(idx_j) * (diff1 - diff0);
+				else break;
+      }
+      diff0 = diff1;
+    }
   }
   return out;
 }
@@ -171,10 +198,10 @@ arma::rowvec re2(const arma::vec& b,
 //' @noRd
 // [[Rcpp::export(rng = false)]]
 arma::vec reGehan(const arma::vec& a,
-		  const arma::mat& X,
-		  const arma::vec& T,
-		  const arma::vec& Y,
-		  const arma::vec& W) {
+									const arma::mat& X,
+									const arma::vec& T,
+									const arma::vec& Y,
+									const arma::vec& W) {
   arma::uword const n = Y.n_elem;
   arma::uword const p = a.n_elem;
   arma::vec out(p, arma::fill::zeros);
@@ -187,12 +214,18 @@ arma::vec reGehan(const arma::vec& a,
   };
   std::set<cmp_par, decltype(cmp)> indices(cmp);
   double w_sum{};
+	double w_sum0{};
+	double w_sum1{};
   arma::vec x_col_sum(p, arma::fill::zeros);
+  arma::vec x_col_sum0(p, arma::fill::zeros);
+  arma::vec x_col_sum1(p, arma::fill::zeros);
   {
     auto const idx_i = idx[0];
     indices.emplace(yexa[idx_i], idx_i);
     x_col_sum += W(idx_i) * X.col(idx_i);
     w_sum += W(idx_i);
+		w_sum0 = w_sum;
+		x_col_sum0 = x_col_sum;
   }
   auto indices_head = indices.begin();
   for(arma::uword i = 1; i < n; ++i) {
@@ -209,6 +242,17 @@ arma::vec reGehan(const arma::vec& a,
     x_col_sum += W(idx_i) * X.col(idx_i);
     w_sum += W(idx_i);
     out += W(idx_i) * (w_sum * X.col(idx_i) - x_col_sum);
+		w_sum1 = w_sum;
+		x_col_sum1 = x_col_sum;
+		// adjust for ties
+		for(arma::uword j = 1; j <= i; ++j) {
+			auto const idx_j = idx[i - j];
+			if (texa[idx_i] == texa[idx_j])
+				out +=  W(idx_j) * ((w_sum1 - w_sum0) * X.col(idx_j) - (x_col_sum1 -x_col_sum0));
+			else break;
+		}
+		w_sum0 = w_sum1;
+		x_col_sum0 = x_col_sum1;
   }
   return out;
 }
@@ -319,6 +363,11 @@ arma::rowvec am1(const arma::vec& a,
       --indices_head;
     w_sum += 1; //W(idx_i);
     de(idx_i) = w_sum;
+		for(arma::uword j = 1; j <= i; ++j) {
+      if (T[idx_i] == T(idx[i - j]))
+				de(idx[i - j]) = w_sum;
+      else break;
+    }
   }
   for (arma::uword k = 0; k < n; k++) {
     for (arma::uword i = 0; i < nm; i++) {
@@ -489,25 +538,50 @@ arma::rowvec temLog(const arma::vec& a,
 }
 
 // [[Rcpp::export(rng = false)]]
-Rcpp::NumericVector temGehan(const arma::vec& a,
-			     const arma::vec& b,
-			     const arma::mat& X,
-			     const arma::vec& Y,
-			     const arma::vec& Z,
-			     const arma::vec& D,
-			     const arma::vec& W) {
-  int n = Y.n_elem;
-  int p = a.n_elem;
-  Rcpp::NumericVector out(p);
-  arma::mat ebax = repmat(D, 1, n) % repmat(exp(X * (b - a)) % Z, 1, n).t();
-  arma::vec yexa = Y % exp(X * a);
-  arma::mat Iij = arma::conv_to<arma::mat>::from(repmat(yexa, 1, n) <= repmat(yexa, 1, n).t());
-  for (int i = 0; i < p; i++) {
-    arma::mat xdif=arma::conv_to<arma::mat>::from(repmat(W % X.col(i), 1, n) -
-						  repmat(W % X.col(i), 1, n).t()); 
-    out[i] = accu(xdif % ebax % Iij);
+arma::vec temGehan(const arma::vec& a,
+		   const arma::vec& b,
+		   const arma::mat& X,
+		   const arma::vec& Y,
+		   const arma::vec& Z,
+		   const arma::vec& D,
+		   const arma::vec& W) {
+  arma::uword const n = Y.n_elem;
+  arma::uword const p = a.n_elem;
+  arma::vec out(p, arma::fill::zeros);
+  if(n < 1) return out;
+  arma::vec yexa = log(Y) + X.t() * a;
+  yexa.replace(-arma::datum::inf, arma::datum::nan);
+  yexa.replace(arma::datum::nan, yexa.min());
+  arma::vec ebaxZ = Z % exp(X.t() * (b - a));
+  arma::uvec const idx = arma::sort_index(yexa);
+  auto cmp = [](cmp_par const &x, cmp_par const &y){
+    return x.first <= y.first;
+  };
+  std::set<cmp_par, decltype(cmp)> indices(cmp);
+  double w_sum{};
+  arma::vec x_col_sum(p, arma::fill::zeros);
+  {
+    auto const idx_i = idx[0];
+    indices.emplace(yexa[idx_i], idx_i);
+    x_col_sum = sum(matvec2(X, W % ebaxZ), 1);
+    w_sum = sum(W % ebaxZ);
+    out = D(idx_i) * W(idx_i) * (w_sum * X.col(idx_i) - x_col_sum);
+  }
+  auto indices_head = indices.begin();
+  for(arma::uword i = 1; i < n; ++i) {
+    auto const idx_i = idx[i];
+    indices.emplace(yexa[idx_i], idx_i);
+    if(yexa[idx_i] > indices_head->first) {
+      while(yexa[idx_i] > indices_head->first) {
+	x_col_sum -= W(indices_head->second) * ebaxZ(indices_head->second) * X.col(indices_head->second);
+	w_sum -= W(indices_head->second) * ebaxZ(indices_head->second);
+	++indices_head;
+      }
+    }
+    else --indices_head;
+    if (w_sum > 0)
+      out += D(idx_i) * W(idx_i) * (w_sum * X.col(idx_i) - x_col_sum);
   }
   return out / n / n;
 }
 
-// Resampling 
